#define _GNU_SOURCE

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <ctype.h>

#include "conf.h"
#include "key.h"

static char error[256];

char *key_error(void)
{
	return error;
}

/*
 * Validates the password
 */
int key_pass_validate(const char *pass, const char *pass2)
{
	unsigned char len, i;

	if (strcmp(pass, pass2) != 0) {
		snprintf(error, sizeof(error),
			"passwords do not match");
		return -1;
	}

	len = strlen(pass);

	if (len % 2 != 0) {
		snprintf(error, sizeof(error),
			"password must have an even"
			" number of characters");
		return -1;
	}

	if (len < cfg_min_pass_len) {
		snprintf(error, sizeof(error),
			"password too short (minimum %hhu)",
			cfg_min_pass_len);
		return -1;
	}

	for (i = 0; i < len; i++) {
		unsigned char c;

		c = toupper(pass[i]);
		if ((c <= '0') || (c > 'F'))
			break;
	}
	if (i < len) {
		snprintf(error, sizeof(error),
			"invalid chars found (0x%02hhx)", pass[i]);
		return -1;
	}

	return 0;
}

/*
 * Save a key to file
 * Returns -1 on error
 */
int key_save(const unsigned int uid, const struct key *k)
{
	int ret, fd;

	ret = -1;
	while (1) {
		char path[256], path2[256];
		ssize_t n, len;
		char buf[4096];
		struct timeval now;

		snprintf(path, sizeof(path), "%s/%u", KEY_DIR, uid);
		mkdir(path, 0700);

		snprintf(path, sizeof(path), "%s/%u/%03hhu.tmp",
			KEY_DIR, uid, k->id);

		fd = open(path, O_WRONLY | O_CREAT | O_EXCL | O_TRUNC, 0400);
		if (fd == -1) {
			snprintf(error, sizeof(error),
				"cannot open key file [%s]: %s",
				path, strerror(errno));
			break;
		}

		gettimeofday(&now, NULL);

		snprintf(buf, sizeof(buf), "%s\n%s\n%ld\n%s\n%ld\n%s\n",
			k->name, k->key, k->ts, k->ip, k->itime, k->pass);

		len = strlen(buf);
		n = write(fd, buf, len);
		if (n != len) {
			snprintf(error, sizeof(error),
				"cannot write to tmp file %s: %s",
				path, strerror(errno));
			break;
		}

		n = fsync(fd);
		if (n == -1) {
			snprintf(error, sizeof(error),
				"cannot fsync tmp file: %s",
				strerror(errno));
			break;
		}

		n = close(fd);
		if (n == -1) {
			snprintf(error, sizeof(error),
				"cannot close tmp file: %s",
				strerror(errno));
			break;
		}
		fd = -1;

		snprintf(path2, sizeof(path2), "%s/%u/%03hhu",
			KEY_DIR, uid, k->id);
		n = rename(path, path2);
		if (n == -1) {
			snprintf(error, sizeof(error),
				"cannot rename tmp file: %s",
				strerror(errno));
			break;
		}

		ret = 0;
		break;
	}
	if (fd != -1)
		close(fd);

	return ret;
}

/*
 * Loads a key from a file
 * Format is like this:
 *	name\n
 *	key\n
 *	ts\n
 *	ip\n
 *	itime\n
 *	pass\n
 * Returns -1 on error.
 */
int key_load(struct key *k, const unsigned int uid, const unsigned char key_id)
{
	int ret, fd;

	ret = -1;
	while (1) {
		char path[256], buf[256], sts[16];
		ssize_t n;
		unsigned short pos;
		unsigned char i;

		snprintf(path, sizeof(path), "%s/%u/%03hhu",
			KEY_DIR, uid, key_id);

		fd = open(path, O_RDONLY);
		if (fd == -1) {
			snprintf(error, sizeof(error),
				"cannot open key file [%s]: %s",
				path, strerror(errno));
			break;
		}

		n = read(fd, buf, sizeof(buf));
		if (n == -1) {
			snprintf(error, sizeof(error),
				"cannot read from file %s: %s",
				path, strerror(errno));
			break;
		}

		memset(k, 0, sizeof(struct key));

		pos = 0;
		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			k->name[i++] = buf[pos++];
		k->name[i] = '\0'; pos++;

		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			k->key[i++] = buf[pos++];
		k->key[i] = '\0'; pos++;

		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			sts[i++] = buf[pos++];
		sts[i] = '\0';
		k->ts = strtol(sts, NULL, 10); pos++;

		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			k->ip[i++] = buf[pos++];
		k->ip[i] = '\0'; pos++;

		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			sts[i++] = buf[pos++];
		sts[i] = '\0';
		k->itime = strtol(sts, NULL, 10); pos++;

		i = 0;
		while ((pos < n) && (buf[pos] != '\n'))
			k->pass[i++] = buf[pos++];
		k->pass[i] = '\0'; pos++;

		k->id = key_id;
		ret = 0;
		break;
	}
	if (fd != -1)
		close(fd);

	return ret;
}

/*
 * Removes a key
 */
int key_remove(const unsigned int uid, const unsigned char id)
{
	char path[256];
	struct stat s;
	int r;

	snprintf(path, sizeof(path), "%s/%u/%03hhu", KEY_DIR, uid, id);

	r = stat(path, &s);
	if (r == -1) {
		snprintf(error, sizeof(error),
			"key id not found");
		return -1;
	}

	r = unlink(path);
	if (r == -1) {
		snprintf(error, sizeof(error),
			"cannot remove key: %s", strerror(errno));
		return -1;
	}

	return 0;
}
