/*
 * Based on http://www.netfilter.org/projects/libnetfilter_queue/doxygen/html/nf-queue_8c_source.html
 * Based on https://code.woboq.org/linux/linux/net/netfilter/nfnetlink_queue.c.html
 * Based on libmnl example.
 * Author: Catalin(ux) M. BOIE
 */

// TODO: Cache codes already computed!
// TODO: What ports to open? Based on user/group?
// TODO: If the IP is validated, do not test anymore anything.
// TODO: We need to specify the time after we request to re-authenticate.
// TODO: Log to syslog all operations.

#define _GNU_SOURCE

#define KEY_DIGITS	16

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/epoll.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/capability.h>
#include <grp.h>
#include <pwd.h>
#include <ctype.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <linux/if_ether.h>

#include <libmnl/libmnl.h>
#include <linux/netfilter.h>
#include <linux/netfilter/nfnetlink.h>

#include <linux/types.h>
#include <linux/netfilter/nfnetlink_queue.h>

#include <libnetfilter_queue/libnetfilter_queue.h>

#include <linux/netfilter/nfnetlink_conntrack.h>


#include "nf2fa_common.h"
#include "util.h"
#include "conf.h"
#include "totp.h"
#include "key.h"
#include "protocol.h"
#include "brute.h"

/* Keep track of IP and expiration */
#define MAX_IPS	1024
struct my_ip
{
	unsigned char	src[16];
	unsigned char	src_len;
	unsigned char	pad[3];
	unsigned int	expire;	/* when we should invalidate this IP */
};
static struct my_ip my_ips[MAX_IPS];

/* Here we keep track of 2fa codes */
struct user
{
	unsigned int	uid;
	unsigned char	pad[4];
	struct key	*head, *tail;
	struct user	*next;
};
static struct user *users_head, *users_tail;

static struct mnl_socket *nl;

static struct nlmsghdr *nfq_hdr_put(unsigned char *buf, const int type,
	const uint32_t queue_num)
{
	struct nlmsghdr *nlh = mnl_nlmsg_put_header(buf);
	nlh->nlmsg_type = (NFNL_SUBSYS_QUEUE << 8) | type;
	nlh->nlmsg_flags = NLM_F_REQUEST;

	struct nfgenmsg *nfg = mnl_nlmsg_put_extra_header(nlh, sizeof(*nfg));
	nfg->nfgen_family = AF_UNSPEC;
	nfg->version = NFNETLINK_V0;
	nfg->res_id = htons(queue_num);

	return nlh;
}

static int send_verdict(const int queue_num, const uint32_t id,
	const int verdict, const unsigned int nfmark,
	const unsigned int ctmark)
{
	unsigned char buf[MNL_SOCKET_BUFFER_SIZE];
	struct nlmsghdr *nlh;
	struct nlattr *nest;

	nlh = nfq_hdr_put(buf, NFQNL_MSG_VERDICT, queue_num);
	nfq_nlmsg_verdict_put(nlh, id, verdict);

	if (nfmark)
		nfq_nlmsg_verdict_put_mark(nlh, htonl(nfmark));

	if (ctmark) {
		/* Set the connmark. First, start NFQA_CT section: */
		nest = mnl_attr_nest_start(nlh, NFQA_CT);
		if (!nest) {
			xlog("%s: mnl_attr_nest_start: %s\n", __func__, strerror(errno));
			return -1;
		}

		/* Then, add the connmark attribute: */
		mnl_attr_put_u32(nlh, CTA_MARK, htonl(ctmark));
		/* TODO: it is not working because Fedora kernel, at least
		 * does not enable needed config option *_GLUE_CT */
		/* Seems Debian is enabling it! */

		/* more conntrack attributes, e.g. CTA_LABEL/ZONE, could be set here */

		/* end conntrack section */
		mnl_attr_nest_end(nlh, nest);
	}

	if (mnl_socket_sendto(nl, nlh, nlh->nlmsg_len) < 0) {
		xlog("%s: mnl_socket_send: %s\n", __func__, strerror(errno));
		return -1;
	}

	return 0;
}

/*
 * Try to find the code inside the packet
 * Returns 0 if not ok, else, the command
 */
static int test_2fa_bin(const unsigned char *data, const unsigned int data_len,
	const unsigned int now, const char *ip)
{
	unsigned short i, j, start, last_check = 0xFFFF;
	unsigned int now30;
	unsigned char cmd;
	struct user *q;
	struct key *k;
	char pin[7];
	int r;

	if (data_len == 0)
		return 0;

	xlog("DEBUG: %s: %s data_len=%u\n", __func__, ip, data_len);
	dump(data, data_len);

	// Find first 0xAA
	start = 0;
	while ((start < data_len) && (data[start] != 0xaa))
		start++;

	now30 = now / 30;

	q = users_head;
	while (q) {
		k = q->head;
		while (k) {
			unsigned char pass_len, needed;

			xlog("DEBUG: uid %u key %03hhu\n", q->uid, k->id);

			pass_len = strlen(k->pass) / 2;
			needed = 1 + pass_len + 3 + 1; // 0xaa + pass + pin + cmd

			i = start;
			while (1) {
				char pass[15];
				unsigned short off;

				if (i + needed > data_len) {
					//xlog("\tDEBUG: too little data: i(%u) + needed(%hhu) > data_len(%hu)\n",
					//	i, needed, data_len);
					break;
				}

				if (data[i] != 0xaa) {
					i++;
					continue;
				}

				// Optimization - skip checking same string
				if ((last_check != 0xFFFF)
					&& (memcmp(data + last_check, data + i, needed) == 0)) {
					i += needed;
					continue;
				}
				last_check = i;

				off = i + 1; // skip 0xaa

				// Check if password is correct
				for (j = 0; j < pass_len; j++)
					sprintf(pass + j * 2, "%02x", data[off + j]);
				off += pass_len;

				if (strcasecmp(k->pass, pass) != 0) {
					xlog("\tDEBUG: passwords do not match [%s] != [%s]\n", k->pass, pass);
					i++;
					continue;
				}
				xlog("\tDEBUG: passwords match!\n");

				// Check pin
				snprintf(pin, sizeof(pin), "%02hhx%02hhx%02hhx",
					data[off], data[off + 1], data[off + 2]);
				off += 3;
				//xlog("\tDEBUG: Packet pin: %s\n", pin);

				r = totp_verify(k->key, now30, k->ts / 30, pin);
				if (r != 1) {
					xlog("\tDEBUG: pin does not match\n");
					i++;
					continue;
				}
				xlog("\tDEBUG: pin match\n");

				cmd = data[off];
				xlog("\tDEBUG: cmd=0x%02hhx\n", cmd);

				k->ts = now;

				switch (cmd) {
				case 0x11: // open
					snprintf(k->ip, sizeof(k->ip), "%s", ip);
					// TODO: maybe we should not allow access if we cannot save key? But it is saved in memory!
					xlog("%s: uid %u: key %03hhu: authorized\n",
						ip, q->uid, k->id);
					break;

				case 0xcc: // close
					xlog("%s: uid %u: key %03hhu: closed the firewall\n",
						ip, q->uid, k->id);
					break;

				default:
					return cmd;
				}

				key_save(q->uid, k);
				return cmd;
			}
			k = k->next;
		}
		q = q->next;
	}

	return 0;
}

/*
 * Binary version
 */
static int test_2fa(const unsigned char *data, const unsigned int data_len,
	const unsigned int now, const char *ip)
{
	int r;
	unsigned char buf[4096];
	unsigned short i, j, max;

	// This test is for ICMP packets
	r = test_2fa_bin(data, data_len, now, ip);
	if (r > 0)
		return r;

	// In UDP packets, we have 'aa' as 0x61 0x61.
	// We try to find this string and test from there.
	// We will not try packets odd aligned.

	max = sizeof(buf);
	if (data_len / 2 < max)
		max = data_len / 2;

	// Find first marker
	for (i = 0; i < max; i += 2) {
		if ((data[i] != 0x61) || (data[i + 1] != 0x61))
			continue;

		xlog("\tDEBUG: Found 0x61 0x61 marker at offset %hu\n", i);

		// we found a start marker - let's see how much we can use
		j = 0;
		while (1) {
			unsigned char c;

			c = data[i];
			//xlog("\tDEBUG: i=%hu j=%hu c=0x%02hhx\n", i, j, c);
			if ((c >= '0') && (c <= '9')) {
				if (i % 2 == 0) {
					buf[j] = (c - '0') << 4;
				} else {
					buf[j] |= (c - '0');
					j++;
				}
				i++;
				continue;
			}

			c = toupper(c);
			if ((c >= 'A') && (c <= 'F')) {
				if (i % 2 == 0) {
					buf[j] = (10 + (c - 'A')) << 4;
				} else {
					buf[j] |= 10 + (c - 'A');
					j++;
				}
				i++;
				continue;
			}

			break;
		}

		r = test_2fa_bin(buf, j, now, ip);
		if (r > 0)
			return r;

		if (i % 2 != 0)
			i++;
	}

	return 0;
}

/*
 * Check if IP is in the list (1 if found)
 * Returns 1 if ok
 * Returns 0 if not found
 */
static int check_ip(const unsigned char *src, const unsigned int src_len,
	unsigned int *pos, const struct timeval *now)
{
	unsigned int i;

	*pos = 0xFFFFFFFF; // this means that we did not find it

	for (i = 0; i < MAX_IPS; i++) {
		if (my_ips[i].src_len != src_len)
			continue;

		if (memcmp(my_ips[i].src, src, src_len) != 0)
			continue;

		// Found it
		*pos = i;

		if (my_ips[i].expire < now->tv_sec)
			return 0;

		return 1;
	}

	return 0;
}

/*
 * Add IP to the list
 * @pos - Position where we shoud store the new entry
 */
static void add_ip(const unsigned int pos, const unsigned char *src,
	const unsigned char src_len, const time_t now)
{
	unsigned int j;

	j = pos;
	if (j == 0xFFFFFFFF) {
		unsigned int i;
		// Find a free/expired slot
		for (i = 0; i < MAX_IPS; i++) {
			if (my_ips[i].src_len == 0) {
				j = i;
				break;
			}

			if (my_ips[i].expire > now) {
				j = i;
				break;
			}
		}
	}

	if (j == 0xFFFFFFFF) {
		xlog("We could not find a free slot"
			"; the accept will not be permanent.\n");
		return;
	}

	memcpy(my_ips[j].src, src, src_len);
	my_ips[j].src_len = src_len;
	my_ips[j].expire = now + cfg_grant_time;
}

/*
 * Process packet
 */
static int process(const unsigned short eth,
	const unsigned char *p, const unsigned int len,
	unsigned int *nfmark, unsigned int *ctmark)
{
	struct iphdr *iph;
	const unsigned char *data;
	unsigned char src[16];
	unsigned int data_len, src_len;
	struct timeval now;
	unsigned int pos = 0xFFFFFFFF;
	int r, af;
	char ip[40];

	*nfmark = 0;
	*ctmark = 0;

	switch (eth) {
	case ETH_P_IP:
		iph = (struct iphdr *) p;
		data = p + iph->ihl * 4;
		data_len = len - (data - p);
		memcpy(src, p + 12, 4); src_len = 4;
		af = AF_INET;
		break;

	case ETH_P_IPV6:
		data = p + 40;
		data_len = len - (data - p);
		memcpy(src, p + 8, 16); src_len = 16;
		af = AF_INET6;
		break;

	default: return NF_DROP;
	}

	//xlog("\tDEBUG: src: "); dump(src, src_len);
	//xlog("\tDEBUG: packet: "); dump(data, data_len);
	inet_ntop(af, src, ip, sizeof(ip));
	//xlog("\tDEBUG: ip=%s\n", ip);

	gettimeofday(&now, NULL);

	// Checking if the IP is already in the whitelist
	r = check_ip(src, src_len, &pos, &now);
	if (r == 1) {
		// IP is in the whilelist, but we cannot just accept it
		// Because we may have a "close firewall" command.

	  	r = test_2fa(data, data_len, now.tv_sec, ip);
		if (r == 0xcc) {
			xlog("\t%s: User removed from list.\n", ip);
			my_ips[pos].src_len = 0;
			return NF_ACCEPT;
		}

		if (r == 0x11) {
			xlog("\t%s: Extend access because already in the list!\n", ip);
			my_ips[pos].expire = now.tv_sec + cfg_grant_time;
			return NF_ACCEPT;
		}

		xlog("\t%s: in the white list, allow it [nfmark 0x%08x]"
			" [ctmark 0x%08x]\n",
			ip, cfg_nfmark, cfg_ctmark);
		*nfmark = cfg_nfmark;
		*ctmark = cfg_ctmark;
		return NF_REPEAT;
	}

	r = brute_check(src, src_len, now.tv_sec);
	if (r == -1) {
		//xlog("\t%s: User tried too early! Brute force attack?\n", ip);
		return NF_DROP;
	}

	xlog("\t%s: ip is NOT in the white list.\n", ip);
  	r = test_2fa(data, data_len, now.tv_sec, ip);
	if (r == 0x11) {
		add_ip(pos, src, src_len, now.tv_sec);
		return NF_ACCEPT;
	}

	xlog("\t%s: DROP.\n", ip);
	return NF_DROP;
}

/*
 * This is called when a packet is received
 */
static int queue_cb(const struct nlmsghdr *nlh, void *data __attribute__ ((unused)))
{
	struct nfqnl_msg_packet_hdr *ph = NULL;
	struct nlattr *attr[NFQA_MAX + 1] = {};
	unsigned int id = 0, nfmark, ctmark;
	struct nfgenmsg *nfg;
	uint16_t plen;
	unsigned char *payload;
	int ret, verdict;

	if (nfq_nlmsg_parse(nlh, attr) < 0) {
		xlog("Problems parsing netlink message!\n");
		return MNL_CB_ERROR;
	}

	nfg = mnl_nlmsg_get_payload(nlh);

	if (attr[NFQA_PACKET_HDR] == NULL) {
		xlog("Packet header not set!\n");
		return MNL_CB_ERROR;
	}

	plen = mnl_attr_get_payload_len(attr[NFQA_PAYLOAD]);
	payload = mnl_attr_get_payload(attr[NFQA_PAYLOAD]);

	ph = mnl_attr_get_payload(attr[NFQA_PACKET_HDR]);
	id = ntohl(ph->packet_id);

	xlog("packet received (id=%u hw=0x%04hx hook=%hhu, payload len %u\n",
		id, ntohs(ph->hw_protocol), ph->hook, plen);

	verdict = process(ntohs(ph->hw_protocol), payload, plen, &nfmark, &ctmark);
	ret = send_verdict(ntohs(nfg->res_id), id, verdict, nfmark, ctmark);
	if (ret == -1)
		return MNL_CB_ERROR;

	return MNL_CB_OK;
}

/*
 * Add a key to a struct user, ordering them correctly
 */
static void key_add(struct user *u, struct key *k)
{
	struct key *q, *prev;

	prev = NULL;
	q = u->head;
	while (q) {
		if (q->id > k->id)
			break;

		prev = q;
		q = q->next;
	}

	if (prev == NULL) {
		k->next = u->head;
		u->head = k;
	} else {
		k->next = prev->next;
		prev->next = k;
	}
}

/*
 * Returns a struct user by uid
 */
static struct user *uid_to_user(const unsigned int uid)
{
	struct user *q;

	q = users_head;
	while (q) {
		if (q->uid == uid)
			return q;

		q = q->next;
	}

	return NULL;
}

/*
 * Returns the next free key slot.
 * Returns 0xFF if not available
 */
static unsigned char key_next_free(struct user *u)
{
	struct key *k;
	unsigned char min = 0;

	k = u->head;
	while (k) {
		if (k->id > min)
			return min;

		k = k->next;
		min++;
	}

	return min;
}

/*
 * Process a client command
 * Returns the number of bytes stored in @buf
 * TODO: do we really heave at least 1 byte?
 */
static int process_client(const unsigned int uid, unsigned char *buf,
	unsigned int n)
{
	unsigned char cmd;
	unsigned short i = 0;
	int r;

	i = 0;
	cmd = buf[i++]; n--;

	while (cmd == NF2FA_CMD_ENROLL) {
		struct user *u;
		struct key k, *k2;
		unsigned char len, len2;
		struct timeval now;

		memset(&k, 0, sizeof(struct key));

		if (n < 1)
			break;
		len = buf[i++]; n--;
		if (len > n)
			break;
		if (len > sizeof(k.name) - 1)
			len2 = sizeof(k.name) - 1;
		else
			len2 = len;
		memcpy(k.name, buf + i, len2);
		k.name[len2] = '\0';
		i += len; n -= len;
		//xlog("DEBUG: ENROLL name=[%s]\n", k.name);

		if (n < 1)
			break;
		len = buf[i++]; n--;
		if (len > n)
			break;
		if (len > sizeof(k.pass) - 1)
			len2 = sizeof(k.pass) - 1;
		else
			len2 = len;
		memcpy(k.pass, buf + i, len2);
		k.pass[len2] = '\0';
		i += len; n -= len;
		//xlog("DEBUG: ENROLL pass=[%s]\n", k.pass);

		i = 1; /* we now overwrite the buffer */

		if (strlen(k.pass) < cfg_min_pass_len)
			return i + protocol_enqueue_error(buf + i,
				"password too short (minimum %hhu)", cfg_min_pass_len);

		u = uid_to_user(uid);
		if (!u)
			k.id = 0;
		else
			k.id = key_next_free(u);
		if (k.id == 0xFF)
			return i + protocol_enqueue_error(buf + i,
				"no free slots; please try to free some");

		r = totp_base64_generate(k.key, KEY_DIGITS);
		if (r != 0)
			return i + protocol_enqueue_error(buf + i, totp_error());

		gettimeofday(&now, NULL);
		k.itime = now.tv_sec;

		r = key_save(uid, &k);
		if (r != 0)
			return i + protocol_enqueue_error(buf + i, key_error());

		// Add the key to memory
		u = users_head;
		while (u) {
			if (u->uid == uid)
				break;
			u = u->next;
		}
		if (!u) {
			u = malloc(sizeof(struct user));
			if (!u)
				return i + protocol_enqueue_error(buf + i,
					"cannot alloc memory");
			memset(u, 0, sizeof(struct user));
			u->uid = uid;
			users_head = u;
		} else {
			users_tail->next = u;
		}
		users_tail = u;

		k2 = malloc(sizeof(struct key));
		if (!k2)
			return i + protocol_enqueue_error(buf + i,
				"cannot alloc memory");
		memcpy(k2, &k, sizeof(struct key));

		key_add(u, k2);

		buf[i++] = 0x00; /* marks the success */
		buf[i++] = k.id;
		i += protocol_enqueue_string(buf + i, k.key);
		return i;
	}

	while (cmd == NF2FA_CMD_LIST) {
		struct user *u;
		struct key *k;

		i = 1; /* we now overwrite the buffer */

		buf[i++] = 0x00; /* OK */

		u = uid_to_user(uid);
		if (!u)
			return i;

		k = u->head;
		while (k) {
			unsigned int x;

			buf[i++] = k->id;
			i += protocol_enqueue_string(buf + i, k->name);
			x = htonl(k->ts);
			memcpy(buf + i, &x, 4); i += 4;
			x = htonl(k->itime);
			memcpy(buf + i, &x, 4); i += 4;
			i += protocol_enqueue_string(buf + i, k->ip);

			k = k->next;
		}

		return i;
	}

	while (cmd == NF2FA_CMD_UNENROLL) {
		struct user *u;
		struct key *k, *kprev;
		int r;
		unsigned char id;

		if (n < 1)
			break;

		id = buf[i++]; n--;

		i = 1; /* we now overwrite the buffer */

		r = key_remove(uid, id);
		if (r != 0)
			return i + protocol_enqueue_error(buf + i, key_error());

		// Remove it from memory
		u = uid_to_user(uid);
		if (!u)
			return i + protocol_enqueue_error(buf + i,
				"no enrollments found");

		kprev = NULL;
		k = u->head;
		while (k) {
			if (k->id != id) {
				kprev = k;
				k = k->next;
				continue;
			}

			if (kprev)
				kprev->next = k->next;
			else
				u->head = k->next;
			free(k);
			break;
		}

		buf[i++] = 0x00; /* marks the success */
		xlog("Unenrolled key %03hhu for uid %u.\n", id, uid);
		return i;
	}

	return i + protocol_enqueue_error(buf + i, "unknown or invalid command");
}

/*
 * Load the keys for a user
 */
static int load_keys_uid(struct user *u, const unsigned int uid)
{
	int err;

	err = -1;
	while (1) {
		DIR *d;
		char path[255];

		snprintf(path, sizeof(path), "%s/%u", KEY_DIR, uid);
		d = opendir(path);
		if (!d) {
			xlog("Cannot open dir %s: %s!\n",
				path, strerror(errno));
			break;
		}

		errno = 0;
		while (1) {
			unsigned char key_id;
			struct key *k;
			struct dirent *de;
			int r;

			de = readdir(d);
			if (!de) {
				if (errno == 0)
					err = 0;
				break;
			}

			if (strlen(de->d_name) != 3)
				continue;

			//xlog("  Loading key %s...\n", de->d_name);
			k = malloc(sizeof(struct key));
			if (!k) {
				xlog("Cannot alloc memory!\n");
				break;
			}

			key_id = strtoul(de->d_name, NULL, 10);

			r = key_load(k, uid, key_id);
			if (r != 0)
				break;

			key_add(u, k);
		}

		closedir(d);
		break;
	}

	return err;
}

/*
 * Load all keys from KEY_DIR
 */
static int load_keys(void)
{
	DIR *d;
	int err, r;

	d = opendir(KEY_DIR);
	if (!d) {
		xlog("Cannot open keys dir: %s!\n", strerror(errno));
		return -1;
	}

	err = -1;
	errno = 0;
	while (1) {
		struct dirent *de;
		struct user *u;
		unsigned int uid;

		de = readdir(d);
		if (!de) {
			if (errno == 0)
				err = 0;
			break;
		}

		if (de->d_name[0] == '.')
			continue;

		//xlog("Loading uid [%s]...\n", de->d_name);

		u = malloc(sizeof(struct user));
		if (!u) {
			xlog("Cannot alloc memory!\n");
			break;
		}
		memset(u, 0, sizeof(struct user));

		if (users_head == NULL)
			users_head = u;
		else
			users_tail->next = u;
		users_tail = u;

		uid = strtoul(de->d_name, NULL, 10);
		r = load_keys_uid(u, uid);
		if (r != 0)
			break;
	}
	closedir(d);

	return err;
}

/*
 * Drop privileges
 * returns -1 on error
 */
static int drop_privileges(void)
{
	int ret, err;
	cap_t caps;
	cap_value_t cap_list[1];

	// Be sure the rights are ok
	chmod(KEY_DIR, 0700);

	chdir("/");

	err = -1;
	while (1) {
		caps = cap_get_proc();
		if (!caps) {
			xlog("Cannot obtain caps: %s!\n", strerror(errno));
			break;
		}

		ret = cap_clear(caps);
		if (ret == -1) {
			xlog("Cannot clear caps: %s!\n", strerror(errno));
			break;
		}

		cap_list[0] = CAP_NET_ADMIN;

		ret = cap_set_flag(caps, CAP_EFFECTIVE, 1, cap_list, CAP_SET);
		if (ret == -1) {
			xlog("Cannot set eff flags: %s!\n", strerror(errno));
			break;
		}

		ret = cap_set_flag(caps, CAP_PERMITTED, 1, cap_list, CAP_SET);
		if (ret == -1) {
			xlog("Cannot set perm flags: %s!\n", strerror(errno));
			break;
		}

		ret = cap_set_proc(caps);
		if (ret == -1) {
			xlog("Cannot set caps: %s!\n", strerror(errno));
			break;
		}

		err = 0;
		break;
	}
	if (caps)
		cap_free(caps);

	return err;
}

int main(int argc, char *argv[])
{
	unsigned char buf[MNL_SOCKET_BUFFER_SIZE];
	struct nlmsghdr *nlh;
	int ret, efd, fd1, fd2;
	unsigned int portid;
	unsigned int i;
	struct epoll_event ev;
	struct sockaddr_un sa_un;
	socklen_t slen;
	char *conf_file;

	xlog("Version: %s\n", NF2FA_VERSION);

	if (argc >= 2)
		conf_file = argv[1];
	else
		conf_file = CONF_FILE;
	xlog("conf_file: %s\n", conf_file);

	ret = conf_load(conf_file);
	if (ret == -1)
		return 1;

	ret = drop_privileges();
	if (ret != 0)
		return 1;

	ret = brute_init(cfg_bad_entries);
	if (ret == -1) {
		xerr("cannot init brute module: %s", brute_error());
		return 1;
	}

	ret = load_keys();
	if (ret == -1)
		return 1;


	nl = mnl_socket_open(NETLINK_NETFILTER);
	if (nl == NULL) {
		perror("mnl_socket_open");
		return 1;
	}

	if (mnl_socket_bind(nl, 0, MNL_SOCKET_AUTOPID) < 0) {
		perror("mnl_socket_bind");
		return 1;
	}

	unlink(cfg_sock_path);
	fd2 = socket(AF_UNIX, SOCK_STREAM, 0);
	if (fd2 == -1) {
		xlog("Cannot create unix socket: %s!\n", strerror(errno));
		return 1;
	}

	memset(&sa_un, 0, sizeof(struct sockaddr_un));
	sa_un.sun_family = AF_UNIX;
	strncpy(sa_un.sun_path, cfg_sock_path, sizeof(sa_un.sun_path) - 1);
	slen = sizeof(struct sockaddr_un);
	ret = bind(fd2, (struct sockaddr *) &sa_un, slen);
	if (ret == -1) {
		xlog("Cannot bind unix socket: %s!\n", strerror(errno));
		return 1;
	}
	chmod(cfg_sock_path, 0666);


	portid = mnl_socket_get_portid(nl);

	nlh = nfq_hdr_put(buf, NFQNL_MSG_CONFIG, cfg_queue);
	nfq_nlmsg_cfg_put_cmd(nlh, AF_INET, NFQNL_CFG_CMD_BIND);
	if (mnl_socket_sendto(nl, nlh, nlh->nlmsg_len) < 0) {
		perror("mnl_socket_send cmd_bind");
		return 1;
	}

	nlh = nfq_hdr_put(buf, NFQNL_MSG_CONFIG, cfg_queue);
	nfq_nlmsg_cfg_put_params(nlh, NFQNL_COPY_PACKET, 0xffff);
	/* TODO
	mnl_attr_put_u32(nlh, NFQA_CFG_FLAGS, htonl(NFQA_CFG_F_CONNTRACK));
	mnl_attr_put_u32(nlh, NFQA_CFG_MASK, htonl(NFQA_CFG_F_CONNTRACK));
	*/
	if (mnl_socket_sendto(nl, nlh, nlh->nlmsg_len) < 0) {
		perror("mnl_socket_send copy_packet");
		return 1;
	}

	#if 0
	TODO
	if (!nfq_set_queue_flags(qh, NFQA_CFG_F_CONNTRACK, NFQA_CFG_F_CONNTRACK)) {
		perror("mnl_socket_send copy_packet");
		return 1;
	}
	#endif

	/* ENOBUFS is signalled to userspace when packets were lost
	 * on kernel side. In most cases, userspace isn't interested
	 * in this information, so turn it off. */
	ret = 1;
	mnl_socket_setsockopt(nl, NETLINK_NO_ENOBUFS, &ret, sizeof(int));


	efd = epoll_create1(0);
	if (efd == -1) {
		xlog("Cannot create epoll socket: %s!\n", strerror(errno));
		return 1;
	}

	fd1 = mnl_socket_get_fd(nl);
	memset(&ev, 0, sizeof(struct epoll_event));
	ev.events = EPOLLIN;
	ev.data.fd = fd1;
	if (epoll_ctl(efd, EPOLL_CTL_ADD, fd1, &ev) == -1) {
		xlog("Cannot add nl socket to epoll: %s!\n",
			strerror(errno));
		return 1;
	}

	// Preparing unix socket for control
	listen(fd2, 16);
	memset(&ev, 0, sizeof(struct epoll_event));
	ev.events = EPOLLIN;
	ev.data.fd = fd2;
	if (epoll_ctl(efd, EPOLL_CTL_ADD, fd2, &ev) == -1) {
		xlog("Cannot add unix socket to epoll: %s!\n",
			strerror(errno));
		return 1;
	}


	for (i = 0; i < MAX_IPS; i++)
		my_ips[i].src_len = 0;

	for (;;) {
		ret = epoll_wait(efd, &ev, 1, -1);
		if (ret == -1) {
			xlog("Cannot wait for epoll: %s!\n",
				strerror(errno));
			return 1;
		}

		if (ev.data.fd == fd1) {
			ret = mnl_socket_recvfrom(nl, buf, sizeof(buf));
			if (ret == -1) {
				if (errno == ENOBUFS) {
					fprintf(stderr, "Some packets were lost. Bad.\n");
					continue;
				}

				perror("error receiving from nl fd");
				return 1;
			}

			ret = mnl_cb_run(buf, ret, 0, portid, queue_cb, NULL);
			if (ret < 0){
				perror("mnl_cb_run");
				return 1;
			}
			continue;
		}

		if (ev.data.fd == fd2) {
			ssize_t n;
			struct ucred ucred;
			int csock;
			struct timeval tv;

			csock = accept(fd2, NULL, NULL);
			if (csock == -1) {
				xlog("Cannot accept: %s!\n", strerror(errno));
				continue;
			}

			slen = sizeof(struct ucred);
			if (getsockopt(csock, SOL_SOCKET, SO_PEERCRED,
				&ucred, &slen) == -1) {
				xlog("Cannot find out uid/gid of peer: %s!\n",
					strerror(errno));
				goto clean;
			}
			xlog("Connection from"
				" pid=%ld, euid=%ld, egid=%ld\n",
				(long) ucred.pid, (long) ucred.uid,
				(long) ucred.gid);

			// Set timeout
			tv.tv_sec = 0; tv.tv_usec = 20000;
			slen = sizeof(struct timeval);
			setsockopt(csock, SOL_SOCKET, SO_RCVTIMEO, &tv, slen);

			n = recv(csock, buf, sizeof(buf), 0);
			if (n == -1) {
				xlog("Cannot receive: %s\n", strerror(errno));
				goto clean;
			}
			//xlog("Data on UNIX socket:\n"); dump(buf, n);

			n = process_client(ucred.uid, buf, n);
			if (n == -1)
				goto clean;

			n = send(csock, buf, n, 0);
			if (n == -1)
				xlog("Cannot send: %s!\n", strerror(errno));

			clean:
			close(csock);
		}
	}

	mnl_socket_close(nl);

	return 0;
}
