#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <openssl/hmac.h>
#include <openssl/evp.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include <qrencode.h>

#include "totp.h"
#include "util.h"

static char *totp_base32_tab = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";
static char error[256];

char *totp_error(void)
{
	return error;
}

/*
 * Generate a secret
 * Store the key into @out; it must have at least @digits + 1 bytes
 */
int totp_base64_generate(char *out, const unsigned char digits)
{
	unsigned char i, ra[256];
	ssize_t r;

	r = my_getrandom(ra, digits, 0);
	if (r == -1) {
		snprintf(error, sizeof(error), "cannot get enough random bytes");
		return -1;
	}

	for (i = 0; i < digits; i++)
		out[i] = totp_base32_tab[ra[i] % 32];
	out[digits] = '\0';

	return 0;
}

/*
 * Returns the index of the char
 */
unsigned char totp_char_to_index(const char c)
{
	const char *x;
	int q;

	q = toupper(c);
	x = strchr(totp_base32_tab, q);
	if (x == NULL) {
		snprintf(error, sizeof(error), "invalid base32 char");
		return 0xFF;
	}

	return x - totp_base32_tab;
}

/*
 * Transforms a base32 google string into a binary array
 * Returns the binary string in @out
 * Returns -1 on error, >0 on success (number of bytes in @out)
 */
int totp_base32_decode(unsigned char *out, const unsigned char out_size,
	const char *s)
{
	unsigned char n, buf_bits;
	unsigned short buf;
	unsigned int i;

	i = 0;
	buf = 0;
	buf_bits = 0;
	while (*s) {
		while (*s && (buf_bits < 8)) {
			n = totp_char_to_index(*s); s++;
			if (n == 0xFF)
				return -1;
			buf = (buf << 5) | n;
			buf_bits += 5;
		}

		// Less than 8 bits left, pad right with 0
		if (buf_bits < 8) {
			buf <<= (8 - buf_bits);
			buf_bits = 8;
		}

		// Now we have more than 8 bits in buffer, extract
		buf_bits -= 8;
		out[i++] = buf >> buf_bits;
		buf &= ((1 << buf_bits) - 1);

		if (i == out_size) {
			snprintf(error, sizeof(error), "buffer too small");
			return -1;
		}
	}

	return i;
}

/*
 * Converts a two digit hex string into a unsigned char
 */
static unsigned char totp_hex2bin(const char *s)
{
	unsigned char i, v[2];

	for (i = 0; i < 2; i++) {
		int c;

		c = toupper(s[i]);
		if ((c >= '0') && (c <= '9'))
			v[i] = c - '0';
		else if ((c >= 'A') && (c <= 'F'))
			v[i] = c + 10 - 'A';
		else
			v[i] = 0;
	}

	return v[0] * 16 + v[1];
}

/*
 * Compute a code based on a key, a timestamp/30 and number of digits
 * Returns -1 on error (probably the key is invalid), 0 on success (@out
 * will contain the digits)
 * Use maximum 10 digits.
 */
int totp_compute(char *out, const char *key, unsigned int tc,
	unsigned char digits)
{
	unsigned char key_bin[64], key_len, btc[8], i, *p, o, number_len;
	int r;
	char stc[17], number[11];
	unsigned int v;
	unsigned long long z;

	//printf("DEBUG: %s: key=%s tc=%u\n", __func__, key, tc);

	if (digits > 10) {
		snprintf(error, sizeof(error), "too many digits");
		return -1;
	}

	r = totp_base32_decode(key_bin, sizeof(key_bin), key);
	if (r == -1)
		return -1;
	key_len = r;

	snprintf(stc, sizeof(stc), "%016x", tc);
	// Transform stc into a binary
	for (i = 0; i < 8; i++)
		btc[i] = totp_hex2bin(stc + i * 2);

	// HMAC: SHA1: data=btc key=key_bin
	p = HMAC(EVP_sha1(), key_bin, key_len, btc, 8, NULL, NULL);

	o = p[19] & 0x0F;
	v = (p[o] << 24) | (p[o + 1] << 16) | (p[o + 2] << 8) | p[o + 3];
	v &= 0x7FFFFFFF;

	z = 1;
	for (i = 0; i < digits; i++)
		z *= 10;

	number_len = snprintf(number, sizeof(number), "%llu", v % z);

	// Prepending zeros
	for (i = 0; i < digits - number_len; i++)
		out[i] = '0';
	out[i] = '\0';

	strcat(out, number);

	return 0;
}

/*
 * Verifies a token based on @tc
 * Returns -1 on error, 0 if they do not match, 1 if they match
 */
int totp_verify_tc(const char *key, const unsigned int tc, const char *pin)
{
	char cpin[16];
	int r;
	unsigned char digits;

	digits = strlen(pin);
	r = totp_compute(cpin, key, tc, digits);
	if (r == -1)
		return -1;

	if (strcmp(cpin, pin) != 0)
		return 0;

	return 1;
}

/*
 * Verifies a token based on @tc, trying 2 in the past and 2 in the future,
 * to allow clock skew.
 * @last_used - we should not allow to reuse the last token
 * Returns -1 on error, 0 if they do not match, else 1
 */
int totp_verify(const char *key, const unsigned int tc,
	const unsigned int last_used, const char *pin)
{
	unsigned int values[5];
	unsigned char i;

	if (*pin == '\0')
		return -1;

	values[0] = tc;
	values[1] = tc - 1;
	values[2] = tc + 1;
	values[3] = tc - 2;
	values[4] = tc + 2;

	for (i = 0; i < 5; i++) {
		if (values[i] <= last_used)
			continue;

		if (totp_verify_tc(key, values[i], pin) == 1)
			return 1;
	}

	return 0;
}

/*
 * Encodes an URL
 * @out must have the 3 * size of @in
 */
static void encode(char *out, const char *in)
{
	const char *allow = "*-._";

	while (*in) {
		if (isalnum(*in) || strchr(allow, *in)) {
			*out = *in;
			out++; in++;
			continue;
		}

		#if 0
		if (*in == ' ') {
			*out = '+';
			out++; in++;
			continue;
		}
		#endif

		snprintf(out, 4, "%%%02hhx", *in);
		out += 3; in++;
	}
	*out = '\0';
}

/*
 * Generates a UTF-8 QR code
 */
int totp_text(const char *secret, const char *name)
{
	char host[128], extra[256];
	QRcode *qr;
	unsigned char i, j, v1, v2;
	unsigned short x, y, limit;
	const char *sym[4] = { "█", "▄", "▀", " " };
	const unsigned char sym_len[4] = { 3, 3, 3, 1 };
	char ename[128], ehost[400];

	gethostname(host, sizeof(host) - 1);
	host[sizeof(host) - 1] = '\0';
	encode(ehost, host);

	encode(ename, name);
	snprintf(extra, sizeof(extra), "otpauth://totp/%s?secret=%s&issuer=%s",
		ename, secret, ehost);

	qr = QRcode_encodeString(extra, 0, QR_ECLEVEL_L, QR_MODE_8, 0 /*case*/);
	if (!qr) {
		snprintf(error, sizeof(error),
			"cannot generate qr code: %s", strerror(errno));
		return -1;
	}

	// top margin
	for (j = 0; j < qr->width + 4; j++)
		write(STDOUT_FILENO, sym[0], sym_len[0]);
	write(STDOUT_FILENO, "\n", 1);

	limit = qr->width * qr->width;
	for (i = 0; i < qr->width; i += 2) {
		write(STDOUT_FILENO, sym[0], sym_len[0]);
		write(STDOUT_FILENO, sym[0], sym_len[0]);

		for (j = 0; j < qr->width; j++) {
			x = i * qr->width + j;
			y = (i + 1) * qr->width + j;
			v1 = qr->data[x] & 1;
			if (y < limit)
				v2 = qr->data[y] & 1;
			else
				v2 = 0;
			write(STDOUT_FILENO, sym[(v2 << 1) | v1], sym_len[(v2 << 1) | v1]);
		}

		write(STDOUT_FILENO, sym[0], sym_len[0]);
		write(STDOUT_FILENO, sym[0], sym_len[0]);
		write(STDOUT_FILENO, "\n", 1);
	}

	// bottom margin
	for (j = 0; j < qr->width + 4; j++)
		write(STDOUT_FILENO, sym[0], sym_len[0]);
	write(STDOUT_FILENO, "\n", 1);

	QRcode_free(qr);

	return 0;
}

